@@ -2,9 +2,10 @@ import pkg_resources
 
 name = "nb2workflow"
 
-def version():
+def version(print_it=True):
     v = pkg_resources.get_distribution("nb2workflow").version
-    print(v)
+    if print_it:
+        print(v)
 
     return v
 
