@@ -2,7 +2,6 @@ import os
 import json
 import logging
 import pytest
-import tempfile
 
 
 # this can be also set in pytest call
@@ -73,9 +72,9 @@ def test_nbadapter(test_notebook, morph_notebook, caplog):
 
 def test_find_notebooks(caplog):
     from nb2workflow.nbadapter import find_notebooks, NotebookAdapter
-    
-    nb_dir = 'tests/testfiles'
-    single_nb = 'tests/testfiles/lightcurve.ipynb'
+    testfiles_path = os.path.join(os.path.dirname(__file__), 'testfiles')
+    nb_dir = testfiles_path
+    single_nb = os.path.join(testfiles_path, 'lightcurve.ipynb')
     
     nbas = find_notebooks(single_nb)
     assert len(nbas) == 1
@@ -85,7 +84,7 @@ def test_find_notebooks(caplog):
     assert 'Ignoring pattern.' in caplog.text
     
     nbas = find_notebooks(nb_dir)
-    assert len(nbas) == 4
+    assert len(nbas) == 9
     
     nbas = find_notebooks(nb_dir, pattern=r'.*bool')
     assert len(nbas) == 1
@@ -199,3 +198,48 @@ def test_denumpyfy():
 
     r = json.dumps(denumpyfy(data))
     print(r)
+
+
+def test_multiline_parameters():
+    from nb2workflow.nbadapter import NotebookAdapter
+
+    nba = NotebookAdapter('tests/testfiles/multiline.ipynb')
+
+    pars = nba.input_parameters
+
+    assert pars['mline']['python_type'] == dict
+    assert pars['mline']['default_value'] == {'foo': ['bar', 'baz'],
+                                              'spam': ['ham', 'eggs']}
+    assert pars['mline']['owl_type'] == "http://odahub.io/ontology#StructuredParameter"
+    assert pars['mline']['is_optional'] == False
+
+
+    assert pars['opt']['python_type'] == float
+    assert pars['opt']['default_value'] == None
+    assert pars['opt']['owl_type'] == "http://odahub.io/ontology#Float"
+    assert pars['opt']['is_optional'] == True
+
+
+    assert pars['inten']['python_type'] == int
+    assert pars['inten']['default_value'] == 45
+    assert pars['inten']['owl_type'] == "http://odahub.io/ontology#Energy"
+    assert pars['inten']['is_optional'] == False
+
+    assert pars['intfloat']['python_type'] == float
+    assert pars['intfloat']['default_value'] == 10
+    assert pars['intfloat']['owl_type'] == "http://odahub.io/ontology#Float"
+    assert pars['intfloat']['is_optional'] == False
+
+    assert pars['string_param']['python_type'] == str
+    assert pars['string_param']['default_value'] == 'Foo Bar\nContains = Symbol\nSpam Ham\n'
+    assert pars['string_param']['owl_type'] == "http://odahub.io/ontology#LongString"
+    assert pars['string_param']['is_optional'] == False
+
+    assert pars['flag']['python_type'] == bool
+    assert pars['flag']['default_value'] == True
+    assert pars['flag']['owl_type'] == "http://odahub.io/ontology#Boolean"
+    assert pars['flag']['is_optional'] == False
+
+    outp = nba.extract_output_declarations()
+
+    assert outp['static']['value'] == "Just a static\n            but multiline string"
